# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest

import osmium
from helpers import IDCollector


def test_node_geometry(opl_buffer):
    fp = osmium.FileProcessor(opl_buffer("n23 x3 y4"))\
          .with_filter(osmium.filter.GeoInterfaceFilter())

    for n in fp:
        assert n.__geo_interface__ == \
               dict(type='Feature', properties={},
                    geometry=dict(type='Point',
                                  coordinates=(pytest.approx(3), pytest.approx(4))))
        break
    else:
        assert False


def test_way_geometry(opl_buffer):
    data = """\
            n1 x0.001 y0
            n2 x0.002 y0
            w1 Nn1,n2
           """

    fp = osmium.FileProcessor(opl_buffer(data))\
               .with_locations()\
               .with_filter(osmium.filter.EntityFilter(osmium.osm.WAY))\
               .with_filter(osmium.filter.GeoInterfaceFilter())

    for w in fp:
        assert w.__geo_interface__ == \
            dict(type='Feature', properties={},
                 geometry=dict(type='LineString',
                               coordinates=[[pytest.approx(0.001), pytest.approx(0)],
                                            [pytest.approx(0.002), pytest.approx(0)]]))
        break
    else:
        assert False


def test_area_geometry(opl_buffer):
    data = """\
            n1 x0.001 y0
            n2 x0.002 y0
            n3 x0.001 y0.001
            w1 Nn1,n2,n3
            w2 Nn3,n1
            r1 Ttype=multipolygon Mw1@,w2@
           """

    fp = osmium.FileProcessor(opl_buffer(data))\
               .with_areas()\
               .with_filter(osmium.filter.GeoInterfaceFilter())

    for r in fp:
        print(r)
        if r.is_area():
            assert r.__geo_interface__ == \
                dict(type='Feature', properties={},
                     geometry=dict(type='MultiPolygon',
                                   coordinates=[[[[pytest.approx(0.001), pytest.approx(0)],
                                                  [pytest.approx(0.002), pytest.approx(0)],
                                                  [pytest.approx(0.001), pytest.approx(0.001)],
                                                  [pytest.approx(0.001), pytest.approx(0)]]]]))
            break
    else:
        assert False


def test_area_geometry_without_drop(opl_reader):
    data = """\
            n1 x0.001 y0
            n2 x0.002 y0
            n3
           """

    ids = IDCollector()
    osmium.apply(opl_reader(data), osmium.filter.GeoInterfaceFilter(), ids)

    assert ids.nodes == [1, 2]

    ids = IDCollector()
    osmium.apply(opl_reader(data),
                 osmium.filter.GeoInterfaceFilter(drop_invalid_geometries=False),
                 ids)

    assert ids.nodes == [1, 2, 3]


def test_property_tag_filter(opl_buffer):
    data = """\
           n1 x0.001 y0 Ta=1,b=1,c=1,d=1
           n2 x0.001 y0 Tx=1
           n3 x0.001 y0 Ta=1,c=3
           """

    fp = osmium.FileProcessor(opl_buffer(data))\
               .with_filter(osmium.filter.GeoInterfaceFilter(tags=['a', 'b']))

    count = 0
    for obj in fp:
        count += 1
        for k in obj.__geo_interface__['properties']:
            assert k in ['a', 'b']

    assert count == 3



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest
import uuid

import osmium


def assert_tracker_content(ids, nodes, ways, rels):
    assert len(ids.node_ids()) == len(nodes)
    assert all(n in ids.node_ids() for n in nodes), \
           f"Nodes not found: {[n for n in nodes if n not in ids.node_ids()]}"
    assert len(ids.way_ids()) == len(ways)
    assert all(n in ids.way_ids() for n in ways), \
           f"Ways not found: {[n for n in ways if n not in ids.way_ids()]}"
    assert len(ids.relation_ids()) == len(rels)
    assert all(n in ids.relation_ids() for n in rels), \
           f"Relations not found: {[n for n in rels if n not in ids.relation_ids()]}"


def test_add_node():
    ids = osmium.IdTracker()

    ids.add_node(45)

    assert ids.node_ids().get(45)
    assert len(ids.node_ids()) == 1

    ids.node_ids().unset(45)
    assert ids.node_ids().empty()


def test_add_way():
    ids = osmium.IdTracker()

    ids.add_way(45)

    assert ids.way_ids().get(45)
    assert len(ids.way_ids()) == 1

    ids.way_ids().unset(45)
    assert ids.way_ids().empty()


def test_add_relation():
    ids = osmium.IdTracker()

    ids.add_relation(45)

    assert ids.relation_ids().get(45)
    assert len(ids.relation_ids()) == 1

    ids.relation_ids().unset(45)
    assert ids.relation_ids().empty()


def test_add_references_from_file(opl_buffer):
    data = """\
           w1 Nn1,n2,n3
           w2 Nn100,n2,n101
           r1 Mn1000@,w23@,r1@
           """

    ids = osmium.IdTracker()

    for obj in osmium.FileProcessor(opl_buffer(data)):
        ids.add_references(obj)

    assert len(ids.node_ids()) == 6
    assert all(n in ids.node_ids() for n in (1, 2, 3, 100, 101, 1000))

    assert len(ids.way_ids()) == 1
    assert 23 in ids.way_ids()

    assert len(ids.relation_ids()) == 1
    assert 1 in ids.relation_ids()


def test_add_reference_from_python_way():
    ids = osmium.IdTracker()

    class Way:
        nodes = [5, 7, 23, 1, 5]

    ids.add_references(Way())

    assert len(ids.node_ids()) == 4
    assert all(n in ids.node_ids() for n in (1, 5, 7, 23))


def test_add_reference_from_python_relation():
    ids = osmium.IdTracker()

    class Member:
        type = 'r'
        ref = 100
        role = ''

    class Rel:
        members = [('n', 23, ''),
                   ('w', 3, ''),
                   Member()]

    ids.add_references(Rel())

    assert len(ids.node_ids()) == 1
    assert 23 in ids.node_ids()

    assert len(ids.way_ids()) == 1
    assert 3 in ids.way_ids()

    assert len(ids.relation_ids()) == 1
    assert 100 in ids.relation_ids()


def test_contains_references_in_node(opl_buffer):
    ids = osmium.IdTracker()
    ids.add_node(45)

    for obj in osmium.FileProcessor(opl_buffer('n45')):
        assert not ids.contains_any_references(obj)


def test_contains_references_in_way(opl_buffer):
    ids = osmium.IdTracker()
    ids.add_node(45)

    for obj in osmium.FileProcessor(opl_buffer('w3 Nn12,n45')):
        assert ids.contains_any_references(obj)


def test_contains_references_not_in_way(opl_buffer):
    ids = osmium.IdTracker()
    ids.add_way(3)

    for obj in osmium.FileProcessor(opl_buffer('w3 Nn12,n45')):
        assert not ids.contains_any_references(obj)


def test_contains_references_in_relation(opl_buffer):
    ids = osmium.IdTracker()
    ids.add_node(45)

    for obj in osmium.FileProcessor(opl_buffer('r3 Mn12@,n45@')):
        assert ids.contains_any_references(obj)


def test_contains_references_not_in_relation(opl_buffer):
    ids = osmium.IdTracker()
    ids.add_way(3)

    for obj in osmium.FileProcessor(opl_buffer('r3 Mn12@,n45@')):
        assert not ids.contains_any_references(obj)


REF_SRC = """\
w12 Nn1,n2
w90 Nn10,n11
r2 Mn99@
r10 Mn100@,w90@,r2@
"""


@pytest.mark.parametrize('depth', range(3))
def test_complete_backward_references(tmp_path, depth):
    if depth == 0:
        data_file = osmium.io.FileBuffer(REF_SRC.encode('utf-8'), 'opl')
    else:
        data_file = tmp_path / f"{uuid.uuid4()}.opl"
        data_file.write_text(REF_SRC)

    ids = osmium.IdTracker()
    ids.add_way(12)
    ids.add_relation(10)

    ids.complete_backward_references(data_file, relation_depth=depth)

    if depth == 0:
        assert_tracker_content(ids, nodes=(1, 2), ways=(12,), rels=(10,))
    elif depth == 1:
        assert_tracker_content(ids, nodes=(1, 2, 10, 11, 100), ways=(12, 90), rels=(10, 2))
    elif depth == 2:
        assert_tracker_content(ids, nodes=(1, 2, 10, 11, 99, 100), ways=(12, 90), rels=(10, 2))


@pytest.mark.parametrize('depth', range(-1, 2))
def test_complete_forward_references(tmp_path, depth):
    if depth == 0:
        data_file = osmium.io.FileBuffer(REF_SRC.encode('utf-8'), 'opl')
    else:
        data_file = tmp_path / f"{uuid.uuid4()}.opl"
        data_file.write_text(REF_SRC)

    ids = osmium.IdTracker()
    ids.add_node(1)
    ids.add_node(99)

    ids.complete_forward_references(data_file, relation_depth=depth)

    if depth == -1:
        assert_tracker_content(ids, nodes=(1, 99), ways=(12,), rels=[])
    elif depth == 0:
        assert_tracker_content(ids, nodes=(1, 99), ways=(12,), rels=[2])
    elif depth == 1:
        assert_tracker_content(ids, nodes=(1, 99), ways=(12,), rels=[2, 10])


def test_clear_node_id_set():
    ids = osmium.IdTracker()
    for i in range(1000, 1003):
        ids.add_node(i)

    assert len(ids.node_ids()) == 3

    ids.node_ids().clear()

    assert len(ids.node_ids()) == 0



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.

from ._osmium import (InvalidLocationError as InvalidLocationError,
                      apply as apply,
                      BaseHandler as BaseHandler,
                      BaseFilter as BaseFilter,
                      BufferIterator as BufferIterator,
                      SimpleWriter as SimpleWriter,
                      NodeLocationsForWays as NodeLocationsForWays,
                      OsmFileIterator as OsmFileIterator,
                      IdTracker as IdTracker)
from .helper import (make_simple_handler as make_simple_handler,
                     WriteHandler as WriteHandler,
                     MergeInputReader as MergeInputReader)
from .simple_handler import (SimpleHandler as SimpleHandler)
from .file_processor import (FileProcessor as FileProcessor,
                             zip_processors as zip_processors)
from .back_reference_writer import BackReferenceWriter as BackReferenceWriter
from .forward_reference_writer import ForwardReferenceWriter as ForwardReferenceWriter
import osmium.io
import osmium.osm
import osmium.index
import osmium.geom
import osmium.area
import osmium.filter



"""
This also creates stats over an OSM file, only the file is first read into
a string buffer.

Shows how to use input from strings.
"""
import osmium
import sys
import urllib.request as urlrequest

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: python osm_url_stats.py <osmfile>")
        sys.exit(-1)

    data = urlrequest.urlopen(sys.argv[1]).read()

    counter = {'n': 0, 'w': 0, 'r': 0}

    for o in osmium.FileProcessor(osmium.io.FileBuffer(data, sys.argv[1])):
        counter[o.type_str()] += 1

    print("Nodes: %d" % counter['n'])
    print("Ways: %d" % counter['w'])
    print("Relations: %d" % counter['r'])



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest

import osmium
from helpers import CountingHandler


@pytest.mark.parametrize('ent,cnt', [(osmium.osm.NODE, (1, 0, 0)),
                                     (osmium.osm.NODE | osmium.osm.WAY, (1, 1, 0)),
                                     (osmium.osm.ALL, (1, 1, 1))])
def test_entity_filter_simple(opl_reader, ent, cnt):
    data = """\
           n1 Ttype=node
           w1 Ttype=way
           r1 Ttype=rel
           """

    processed = CountingHandler()

    osmium.apply(opl_reader(data), osmium.filter.EntityFilter(ent), processed)

    assert list(cnt) == processed.counts[:3]



"""
Output an OSM file as geojson.

This demonstrates how to use the GeoJSON factory.
"""
import sys
import json

import osmium

geojsonfab = osmium.geom.GeoJSONFactory()


class GeoJsonWriter(osmium.SimpleHandler):

    def __init__(self):
        super().__init__()
        # write the Geojson header
        print('{"type": "FeatureCollection", "features": [')
        self.first = True

    def finish(self):
        print(']}')

    def node(self, n):
        if n.tags:
            self.print_object(geojsonfab.create_point(n), n.tags)

    def way(self, w):
        if w.tags and not w.is_closed():
            self.print_object(geojsonfab.create_linestring(w), w.tags)

    def area(self, a):
        if a.tags:
            self.print_object(geojsonfab.create_multipolygon(a), a.tags)

    def print_object(self, geojson, tags):
        geom = json.loads(geojson)
        if geom:
            feature = {'type': 'Feature', 'geometry': geom, 'properties': dict(tags)}
            if self.first:
                self.first = False
            else:
                print(',')

            print(json.dumps(feature))


def main(osmfile):
    handler = GeoJsonWriter()

    handler.apply_file(osmfile,
                       filters=[osmium.filter.EmptyTagFilter().enable_for(osmium.osm.NODE)])
    handler.finish()

    return 0


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: python %s <osmfile>" % sys.argv[0])
        sys.exit(-1)
    sys.exit(main(sys.argv[1]))



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.

from osmium.osm.mutable import create_mutable_node, create_mutable_way, create_mutable_relation

from osmium.osm.types import (OSMObject as OSMObject,
                              Node as Node,
                              Way as Way,
                              Relation as Relation,
                              Area as Area,
                              Changeset as Changeset,
                              Tag as Tag,
                              TagList as TagList,
                              NodeRef as NodeRef,
                              NodeRefList as NodeRefList,
                              WayNodeList as WayNodeList,
                              OuterRing as OuterRing,
                              InnerRing as InnerRing,
                              RelationMember as RelationMember,
                              RelationMemberList as RelationMemberList)

from osmium.osm._osm import (Location as Location,
                             Box as Box,
                             osm_entity_bits as osm_entity_bits,
                             NOTHING as NOTHING,
                             NODE as NODE,
                             WAY as WAY,
                             RELATION as RELATION,
                             AREA as AREA,
                             OBJECT as OBJECT,
                             CHANGESET as CHANGESET,
                             ALL as ALL)

setattr(Location, '__repr__',
        lambda loc: f'osmium.osm.Location(x={loc.x!r}, y={loc.y!r})'
                    if loc.valid() else 'osmium.osm.Location()')
setattr(Location, '__str__',
        lambda loc: f'{loc.lon_without_check():.7f}/{loc.lat_without_check():.7f}'
                    if loc.valid() else 'invalid')

setattr(Box, '__repr__',
        lambda b: f"osmium.osm.Box(bottom_left={b.bottom_left!r}, top_right={b.top_right!r})")
setattr(Box, '__str__', lambda b: f'({b.bottom_left!s} {b.top_right!s})')



import osmium
import sys

if len(sys.argv) != 3:
    print("Usage: python create_nodecache.py <osm file> <node cache>")
    exit(-1)

reader = osmium.io.Reader(sys.argv[1], osmium.osm.osm_entity_bits.NODE)

idx = osmium.index.create_map("sparse_file_array," + sys.argv[2])
lh = osmium.NodeLocationsForWays(idx)

osmium.apply(reader, lh)

reader.close()



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest
import osmium


def test_list_types():
    ml = osmium.index.map_types()
    assert isinstance(ml, list)
    assert ml


@pytest.mark.parametrize('use_get', [True, False])
@pytest.mark.parametrize('use_set', [True, False])
def test_set_get(use_set, use_get):
    table = osmium.index.create_map("flex_mem")
    if use_set:
        table.set(4, osmium.osm.Location(3.4, -5.6))
    else:
        table[4] = osmium.osm.Location(3.4, -5.6)
    if use_get:
        loc = table.get(4)
    else:
        loc = table[4]
    assert loc.lon == pytest.approx(3.4)
    assert loc.lat == pytest.approx(-5.6)


def test_get_unset():
    table = osmium.index.create_map("flex_mem")
    with pytest.raises(KeyError):
        table.get(56)


def test_array_get_unset():
    table = osmium.index.create_map("flex_mem")
    with pytest.raises(KeyError):
        table[56]


def test_set_negative():
    table = osmium.index.create_map("flex_mem")
    with pytest.raises(TypeError):
        table.set(-4, osmium.osm.Location(3.4, -5.6))


def test_array_set_negative():
    table = osmium.index.create_map("flex_mem")
    with pytest.raises(TypeError):
        table[-4] = osmium.osm.Location(3.4, -5.6)


def test_used_memory():
    table = osmium.index.create_map("flex_mem")
    table.set(4, osmium.osm.Location(3.4, -5.6))

    assert table.used_memory() > 0


def test_clear():
    table = osmium.index.create_map("flex_mem")
    table.set(593, osmium.osm.Location(0.35, 45.3))
    table.get(593)
    table.clear()
    with pytest.raises(KeyError):
        table.get(593)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
""" Tests for the pyosmium-up-to-date script.
"""
import uuid
import datetime as dt

import pytest
import osmium
from osmium.tools.pyosmium_up_to_date import pyosmium_up_to_date
import osmium.replication.utils as rutil

from helpers import IDCollector

# Choosing a future date here, so we don't run into pyosmium's check for old
# data. If you get caught by this: congratulations, you are maintaining a
# 50-year old test.
REPLICATION_BASE_TIME = dt.datetime(year=2070, month=5, day=6, hour=20, tzinfo=dt.timezone.utc)
REPLICATION_BASE_SEQ = 100
REPLICATION_CURRENT = 140


@pytest.fixture
def replication_server(httpserver):
    def _state(seq):
        seqtime = REPLICATION_BASE_TIME + dt.timedelta(hours=seq - REPLICATION_CURRENT)
        timestamp = seqtime.strftime('%Y-%m-%dT%H\\:%M\\:%SZ')
        return f"sequenceNumber={seq}\ntimestamp={timestamp}\n"

    httpserver.no_handler_status_code = 404
    httpserver.expect_request('/state.txt').respond_with_data(_state(REPLICATION_CURRENT))
    for i in range(REPLICATION_BASE_SEQ, REPLICATION_CURRENT + 1):
        httpserver.expect_request(f'/000/000/{i}.opl')\
                  .respond_with_data(f"r{i} M" + ",".join(f"n{i}@" for i in range(1, 6000)))
        httpserver.expect_request(f'/000/000/{i}.state.txt').respond_with_data(_state(i))

    return httpserver.url_for('')


@pytest.fixture
def runner(replication_server):
    def _run(*args):
        return pyosmium_up_to_date(
            ['--server', replication_server, '--diff-type', 'opl'] + list(map(str, args)))

    return _run


def test_no_output_file(runner):
    with pytest.raises(SystemExit):
        runner()


def test_simple_update_no_windback(runner, test_data):
    outfile = test_data("n1 v1 t2070-05-06T19:30:00Z")

    assert 0 == runner('--wind-back', 0, outfile)

    ids = IDCollector()
    osmium.apply(outfile, ids)

    assert ids.nodes == [1]
    assert ids.relations == list(range(139, REPLICATION_CURRENT + 1))


def test_simple_update_override(runner, test_data):
    outfile = test_data("n1 v1 t2070-05-06T19:30:00Z")

    assert 0 == runner(outfile)

    ids = IDCollector()
    osmium.apply(outfile, ids)

    assert ids.nodes == [1]
    assert ids.relations == list(range(138, REPLICATION_CURRENT + 1))


def test_simple_update_new_file(runner, replication_server, test_data, tmp_path):
    outfile = test_data("n1 v1 t2070-05-06T19:30:00Z")
    newfile = tmp_path / f"{uuid.uuid4()}.pbf"

    assert 0 == runner('-o', str(newfile), outfile)

    ids = IDCollector()
    osmium.apply(outfile, ids)

    assert ids.nodes == [1]
    assert ids.relations == []

    ids = IDCollector()
    osmium.apply(newfile, ids)
    assert ids.nodes == [1]
    assert ids.relations == list(range(138, REPLICATION_CURRENT + 1))

    header = rutil.get_replication_header(newfile)

    assert header.url == replication_server
    assert header.sequence == REPLICATION_CURRENT
    assert header.timestamp == REPLICATION_BASE_TIME


def test_update_sequences(runner, test_data, tmp_path):
    outfile = test_data("n1 v1 t2070-05-05T10:30:00Z")
    newfile = tmp_path / f"{uuid.uuid4()}.pbf"

    assert 0 == runner('--end-id', '110', '-o', str(newfile), outfile)

    ids = IDCollector()
    osmium.apply(newfile, ids)
    assert ids.nodes == [1]
    assert ids.relations == list(range(105, 111))

    header = rutil.get_replication_header(newfile)

    assert header.sequence == 110

    # Note: this test only catches holes, no duplicate application.
    assert 0 == runner(newfile)

    ids = IDCollector()
    osmium.apply(newfile, ids)
    assert ids.nodes == [1]
    assert ids.relations == list(range(105, REPLICATION_CURRENT + 1))

    header = rutil.get_replication_header(newfile)

    assert header.sequence == REPLICATION_CURRENT


@pytest.mark.parametrize('end_id,max_size,actual_end', [(107, None, 107),
                                                        (None, 1, 108),
                                                        (105, 1, 105),
                                                        (110, 1, 108)])
def test_update_with_endid(test_data, runner, end_id, max_size, actual_end):
    outfile = test_data("n1 v1 t2070-05-05T06:30:00Z")

    params = [outfile]
    if end_id is not None:
        params.extend(('--end-id', end_id))
    if max_size is not None:
        params.extend(('-s', max_size))

    assert (0 if end_id == actual_end else 1) == runner(*params)

    ids = IDCollector()
    osmium.apply(outfile, ids)

    assert ids.relations == list(range(101, actual_end + 1))


def test_update_with_enddate(test_data, runner, tmp_path):
    outfile = test_data("n1 v1 t2070-05-05T06:30:00Z")
    newfile = tmp_path / f"{uuid.uuid4()}.pbf"

    assert 0 == runner('-E', '2070-05-05T09:30:00Z', '-o', newfile, outfile)

    header = rutil.get_replication_header(newfile)

    assert header.sequence == 105
    assert header.timestamp == dt.datetime(year=2070, month=5, day=5, hour=9,
                                           tzinfo=dt.timezone.utc)

    ids = IDCollector()
    osmium.apply(newfile, ids)

    assert ids.relations == list(range(101, 106))


def test_change_date_too_old_for_replication_source(test_data, runner, caplog):
    outfile = test_data("n1 v1 t2070-04-05T06:30:00Z")

    assert 3 == runner(outfile)
    assert 'No starting point found' in caplog.text


def test_change_id_too_old_for_replication_source(caplog, tmp_path, runner, replication_server):
    outfile = tmp_path / f"{uuid.uuid4()}.pbf"
    h = osmium.io.Header()
    h.set('osmosis_replication_base_url', replication_server)
    h.set('osmosis_replication_sequence_number', '98')

    with osmium.SimpleWriter(outfile, 4000, h) as w:
        w.add_node({'id': 1})

    assert 3 == runner(outfile)
    assert 'Cannot download state information for ID 98' in caplog.text



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Any, Union
import os
from typing_extensions import Buffer

from typing import overload

from .osm import Box, osm_entity_bits

class File:
    """ A wrapper for an OSM data file.
    """
    @property
    def has_multiple_object_versions(self) -> bool:
        """ True when the file is in a data format which supports
            having multiple versions of the same object in the file.
            This is usually the case with OSM history and diff files.
        """
    @has_multiple_object_versions.setter
    def has_multiple_object_versions(self, value: bool) -> None: ...
    def __init__(self, filename: Union[str, 'os.PathLike[str]'], format: str='') -> None:
        """ Initialise a new file object. Normally the file format of the
            file is guessed from the suffix of the file name. It may
            also be set explicitly using the _format_ parameter.
        """
    def parse_format(self, format: str) -> None:
        """ Set the format of the file from a format string.
        """

class FileBuffer:
    """ A wrapper around a buffer containing OSM data.
    """
    def __init__(self, buf: Buffer, format: str) -> None:
        """ Initialise a new buffer object. _buf_ can be any buffer
            that adheres to the Python buffer protocol. The format of
            the data must be defined in the _format_ parameter.
        """
    @property
    def has_multiple_object_versions(self) -> bool:
        """ True when the file is in a data format which supports
            having multiple versions of the same object in the file.
            This is usually the case with OSM history and diff files.
        """
    @has_multiple_object_versions.setter
    def has_multiple_object_versions(self, value: bool) -> None: ...

    def parse_format(self, format: str) -> None:
        """ Set the format of the file from a format string.
        """

class Header:
    """ File header data with global information about the file.
    """
    @property
    def has_multiple_object_versions(self) -> bool:
        """ True when the file is in a data format which supports
            having multiple versions of the same object in the file.
            This is usually the case with OSM history and diff files.
        """
    @has_multiple_object_versions.setter
    def has_multiple_object_versions(self, value: bool) -> None: ...

    def __init__(self) -> None:
        """ Initiate an empty header.
        """

    def add_box(self, box: Box) -> Header:
        """ Add the given bounding box to the list of bounding boxes
            saved in the header.
        """

    def box(self) -> Box:
        """ Return the bounding box of the data in the file. If no
            such information is available, an invalid box is returned.
        """

    def get(self, key: str, default: str = ...) -> str:
        """ Get the value of header option _key_ or return _default_ if
            there is no header option with that name.
        """

    def set(self, key: str, value: str) -> None:
        """ Set the value of header option _key_ to _value_.
        """

class Reader:
    """ Low-level object for reading data from an OSM file.

        A Reader does not expose functions to process the data it has read
        from the file. Use [apply][osmium.apply] for that purpose.
    """
    def __init__(self, filename: Union[str, 'os.PathLike[str]', FileBuffer, File],
                 types: osm_entity_bits = ...) -> None:
        """ Create a new reader object. The input may either be
            a filename or a [File][osmium.io.File] or
            [FileBuffer][osmium.io.FileBuffer] object. The _types_ parameter
            defines which kinds of objects will be read from the input. Any
            types not present will be skipped completely when reading the
            file. Depending on the type of input, this can save quite a bit
            of time. However, be careful to not skip over types that may
            be referenced by other objects. For example, ways need
            nodes in order to compute their geometry.

            Readers may be used as a context manager. In that case, the
            `close()` function will be called automatically when the
            reader leaves the scope.
        """

    def close(self) -> None:
        """ Close any open file handles and free all resources. The
            Reader is unusuable afterwards.
        """

    def eof(self) -> bool:
        """ Check if the reader has reached the end of the input data.
        """

    def header(self) -> Header:
        """ Return the Header structure containing global information about
            the input. What information is available depends on the format
            of the input data.
        """

    def __enter__(self) -> 'Reader':...
    def __exit__(self, *args: Any) -> None:...


class Writer:
    """ Low-level object for writing OSM data into a file. This class does not
        expose functions for receiving data to be written. Have a look at
        [SimpleWriter][osmium.SimpleWriter] for a higher-level interface
        for writing data.
    """
    def __init__(self, ffile: Union[str, 'os.PathLike[str]', File],
                 header: Header = ...) -> None:
        """ Create a new Writer. The output may either be a simple filename
            or a [File][osmium.io.File] object. A custom [Header][osmium.io.Header]
            object may be given, to customize the global file information that
            is written out. Be aware that not all file formats support writing
            out all header information.
        """

    def close(self) -> int:
        """ Close any open file handles and free all resources. The Writer
            is unusable afterwards.
        """



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
"""
Update an OSM file with changes from a OSM replication server.

Diffs are downloaded and kept in memory. To avoid running out of memory,
the maximum size of diffs that can be downloaded at once is limited
to 1 GB per default. This corresponds to approximately 3 days of update.
The limit can be changed with the --size parameter. However, you should
take into account that processing the files requires additional memory
(about 1GB more).

The starting time is automatically determined from the data in the file.
For PBF files, it is also possible to read and write the replication
information from the osmosis headers. That means that after the first update,
subsequent calls to pyosmium-up-to-date will continue the updates from the same
server exactly where they have left of.

This program can update normal OSM data files as well as OSM history files.
It detects automatically on what type of file it is called.

The program returns 0, if updates have been successfully applied up to
the newest data or no new data was available. It returns 1, if some updates
have been applied but there is still data available on the server (either
because the size limit has been reached or there was a network error which
could not be resolved). Any other error results in a return code larger than 1.
The output file is guaranteed to be unmodified in that case.

Some OSM data sources require a cookie to be sent with the HTTP requests.
pyosmium-up-to-date does not fetch the cookie from these services for you.
However, it can read cookies from a Netscape-style cookie jar file, send these
cookies to the server and will save received cookies to the jar file.
"""
from typing import Any, List
import sys
import traceback
import logging
import http.cookiejar

from argparse import ArgumentParser, RawDescriptionHelpFormatter, ArgumentTypeError
import datetime as dt
from textwrap import dedent as msgfmt
from tempfile import mktemp
import os.path

from ..replication import server as rserv
from ..version import pyosmium_release
from .common import ReplicationStart

log = logging.getLogger()


def update_from_osm_server(start: ReplicationStart, options: Any) -> int:
    """ Update the OSM file using the official OSM servers at
        https://planet.osm.org/replication. This strategy will attempt
        to start with daily updates before going down to minutelies.
        TODO: only updates from hourlies currently implemented.
    """
    start.source = "https://planet.osm.org/replication/hour/"
    return update_from_custom_server(start, options)


def update_from_custom_server(start: ReplicationStart, options: Any) -> int:
    """ Update from a custom URL, simply using the diff sequence as is.
    """
    assert start.source

    with rserv.ReplicationServer(start.source, options.server_diff_type) as svr:
        log.info(f"Using replication service at {start.source}")

        svr.set_request_parameter('timeout', options.socket_timeout or None)

        if options.cookie is not None:
            cookie_jar = http.cookiejar.MozillaCookieJar(options.cookie)
            cookie_jar.load(options.cookie)
            svr.set_request_parameter('cookies', cookie_jar)

        current = svr.get_state_info()
        if current is None:
            log.error("Cannot download state information. Is the replication URL correct?")
            return 3
        log.debug(f"Server is at sequence {current.sequence} ({current.timestamp}).")

        if start.seq_id is not None and start.seq_id >= current.sequence:
            log.info("File is already up to date.")
            return 0

        startseq = start.get_sequence(svr)
        if startseq is None:
            log.error(f"No starting point found for time {start.date} on server {start.source}")
            return 3

        if not options.force_update:
            cmpdate = dt.datetime.now(dt.timezone.utc) - dt.timedelta(days=90)
            cmpdate = cmpdate.replace(tzinfo=dt.timezone.utc)
            if start.date is None or start.date < cmpdate:
                log.error(
                  """The OSM file is more than 3 months old. You should download a
                     more recent file instead of updating. If you really want to
                     update the file, use --force-update-of-old-planet.""")
                return 3

        log.info("Starting download at ID %d (max %f MB)"
                 % (startseq, options.outsize or float('inf')))

        outfile = options.outfile
        infile = options.infile

        if outfile is None:
            fdir, fname = os.path.split(infile)
            if options.tmpdir is not None:
                fdir = options.tmpdir
            ofname = mktemp(suffix='-' + fname, dir=fdir)
        else:
            ofname = outfile

        if options.outsize is not None:
            max_size = options.outsize * 1024
        elif options.end is None:
            max_size = 1024 * 1024
        else:
            max_size = None

        if options.end is None:
            end_id = None
        else:
            end_id = options.end.get_end_sequence(svr)
            if end_id is None:
                log.error("Cannot find the end date/ID on the server.")
                return 1

        try:
            extra_headers = {'generator': 'pyosmium-up-to-date/' + pyosmium_release}
            outseqs = svr.apply_diffs_to_file(infile, ofname, startseq,
                                              max_size=max_size, end_id=end_id,
                                              extra_headers=extra_headers,
                                              outformat=options.outformat)

            if outseqs is None:
                log.info("No new updates found.")
                return 3

            if outfile is None:
                os.replace(ofname, infile)
        finally:
            if outfile is None:
                try:
                    os.remove(ofname)
                except FileNotFoundError:
                    pass

    log.info("Downloaded until %d. Server has data available until %d." % outseqs)

    # save cookies
    if options.cookie:
        cookie_jar.save(options.cookie)

    return 0 if (end_id or outseqs[1]) == outseqs[0] else 1


def compute_start_point(options: Any) -> ReplicationStart:
    start = ReplicationStart.from_osm_file(options.infile, options.ignore_headers)

    if options.server_url is not None:
        if start.source is not None and start.source != options.server_url:
            log.error(msgfmt(f"""
                  You asked to use server URL:
                    {options.server_url}
                  but the referenced OSM file points to replication server:
                    {start.source}
                  If you really mean to overwrite the URL, use --ignore-osmosis-headers."""))
            raise ArgumentTypeError("Source URL doesn't match replication headers.")
        if start.source is None:
            start.source = options.server_url
            start.seq_id = None
            if start.date is None:
                raise ArgumentTypeError("Cannot determine start date for file.")

    if start.seq_id is None:
        assert start.date is not None
        start.date -= dt.timedelta(minutes=options.wind_back)

    return start


def get_arg_parser(from_main: bool = False) -> ArgumentParser:

    parser = ArgumentParser(prog='pyosmium-up-to-date',
                            description=__doc__,
                            usage=None if from_main else 'pyosmium-up-to-date [options] <osm file>',
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-v', dest='loglevel', action='count', default=0,
                        help='Increase verbosity (can be used multiple times).')
    parser.add_argument('infile', metavar='<osm file>', help="OSM file to update")
    parser.add_argument('-o', '--outfile', dest='outfile',
                        help='Name output of file. If missing, the input file '
                             'will be overwritten.')
    parser.add_argument('--format', dest='outformat', metavar='FORMAT',
                        help='Format the data should be saved in. '
                             'Usually determined from file name.')
    parser.add_argument('--server', action='store', dest='server_url',
                        help='Base URL of the replication server. Default: '
                             'https://planet.osm.org/replication/hour/ '
                             '(hourly diffs from osm.org)')
    parser.add_argument('--diff-type', action='store', dest='server_diff_type', default='osc.gz',
                        help='File format used by the replication server (default: osc.gz)')
    parser.add_argument('-s', '--size', dest='outsize', metavar='SIZE', type=int,
                        help='Maximum size of change to apply at once in MB. '
                             'Defaults to 1GB when no end ID or date was given.')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--end-id', dest='end',
                       type=ReplicationStart.from_id, metavar='ID',
                       help='Last sequence ID to download.')
    group.add_argument('-E', '--end-date', dest='end', metavar='DATE',
                       type=ReplicationStart.from_date,
                       help='Do not download diffs later than the given date.')
    parser.add_argument('--tmpdir', dest='tmpdir',
                        help='Directory to use for temporary files. '
                             'Usually the directory of input file is used.')
    parser.add_argument('--ignore-osmosis-headers', dest='ignore_headers',
                        action='store_true',
                        help='Ignore potential replication information in the '
                             'header of the input file and search for the '
                             'newest OSM object in the file instead.')
    parser.add_argument('-b', '--wind-back', dest='wind_back', type=int, default=60,
                        help='Number of minutes to start downloading before '
                             'the newest addition to input data. (Ignored when '
                             'the file contains a sequence ID.) Default: 60')
    parser.add_argument('--force-update-of-old-planet', action='store_true',
                        dest='force_update',
                        help="Apply update even if the input data is really old.")
    parser.add_argument('--cookie', dest='cookie',
                        help='Netscape-style cookie jar file to read cookies from and where '
                             'received cookies will be written to.')
    parser.add_argument('--socket-timeout', dest='socket_timeout', type=int, default=60,
                        help='Set timeout for file downloads.')
    parser.add_argument('--version', action='version',
                        version='pyosmium ' + pyosmium_release)

    return parser


def pyosmium_up_to_date(args: List[str]) -> int:
    options = get_arg_parser(from_main=True).parse_args(args)
    log.setLevel(max(3 - options.loglevel, 0) * 10)

    try:
        start = compute_start_point(options)
    except RuntimeError as e:
        log.error(str(e))
        return 2

    try:
        if start.source is None:
            return update_from_osm_server(start, options)

        return update_from_custom_server(start, options)
    except Exception:
        traceback.print_exc()

    return 254


def main() -> int:
    logging.basicConfig(stream=sys.stderr,
                        format='%(asctime)s %(levelname)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    return pyosmium_up_to_date(sys.argv[1:])



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Optional, Callable, TypeVar, TYPE_CHECKING

from .simple_handler import SimpleHandler
from .osm import Node, Way, Relation, Area, Changeset
from .index import create_map
from ._osmium import SimpleWriter, NodeLocationsForWays

from ._osmium import MergeInputReader as MergeInputReader

if TYPE_CHECKING:
    from ._osmium import HandlerLike

T = TypeVar('T')
HandlerFunc = Optional[Callable[[T], None]]


def make_simple_handler(node: HandlerFunc[Node] = None,
                        way: HandlerFunc[Way] = None,
                        relation: HandlerFunc[Relation] = None,
                        area: HandlerFunc[Area] = None,
                        changeset: HandlerFunc[Changeset] = None) -> SimpleHandler:
    """ _(deprecated)_ Convenience function that creates a
        [SimpleHandler](Handler-Processing.md#osmium.SimpleHandler) from a set of
        callback functions. Each of the parameters takes an optional callable
        that must expect a single positional parameter with the object being
        processed.
    """
    class __HandlerWithCallbacks(SimpleHandler):
        pass

    if node is not None:
        setattr(__HandlerWithCallbacks, "node", staticmethod(node))
    if way is not None:
        setattr(__HandlerWithCallbacks, "way", staticmethod(way))
    if relation is not None:
        setattr(__HandlerWithCallbacks, "relation", staticmethod(relation))
    if area is not None:
        setattr(__HandlerWithCallbacks, "area", staticmethod(area))
    if changeset is not None:
        setattr(__HandlerWithCallbacks, "changeset", staticmethod(changeset))

    return __HandlerWithCallbacks()


# WriteHandler no longer exists. SimpleWriter can now function as a handler.
class WriteHandler(SimpleWriter):
    """ (Deprecated) Handler function that writes all data directly to a file.

        This is now simply an alias for SimpleWriter. Please refer to its
        documentation.
    """

    def __init__(self, filename: str, bufsz: int = 4096*1024, filetype: str = "") -> None:
        super().__init__(filename, bufsz=bufsz, filetype=filetype)


def _merge_apply(self: MergeInputReader, *handlers: 'HandlerLike',
                 idx: str = '', simplify: bool = True) -> None:
    if idx:
        lh = NodeLocationsForWays(create_map(idx))
        lh.ignore_errors()
        self._apply_internal(lh, *handlers, simplify=simplify)

    self._apply_internal(*handlers, simplify=simplify)


MergeInputReader.apply = _merge_apply  # type: ignore[method-assign]



"""
Simple example that counts the number of changes on a replication server
starting from a given timestamp for a maximum of n hours.

Shows how to detect the different kind of modifications.
"""
import osmium as osmium
import sys
import datetime as dt
import osmium.replication.server as rserv


class Stats:

    def __init__(self):
        self.added = 0
        self.modified = 0
        self.deleted = 0

    def add(self, o):
        if o.deleted:
            self.deleted += 1
        elif o.version == 1:
            self.added += 1
        else:
            self.modified += 1

    def outstats(self, prefix):
        print("%s added: %d" % (prefix, self.added))
        print("%s modified: %d" % (prefix, self.modified))
        print("%s deleted: %d" % (prefix, self.deleted))


class FileStatsHandler(osmium.SimpleHandler):
    def __init__(self):
        super(FileStatsHandler, self).__init__()
        self.nodes = Stats()
        self.ways = Stats()
        self.rels = Stats()

    def node(self, n):
        self.nodes.add(n)

    def way(self, w):
        self.ways.add(w)

    def relation(self, r):
        self.rels.add(r)


if __name__ == '__main__':
    if len(sys.argv) != 4:
        print("Usage: python osm_replication_stats.py <server_url> <start_time> <max kB>")
        sys.exit(-1)

    server_url = sys.argv[1]
    start = dt.datetime.strptime(sys.argv[2], "%Y-%m-%dT%H:%M:%SZ")
    start = start.replace(tzinfo=dt.timezone.utc)
    maxkb = min(int(sys.argv[3]), 10 * 1024)

    repserv = rserv.ReplicationServer(server_url)

    seqid = repserv.timestamp_to_sequence(start)
    print("Initial sequence id:", seqid)

    h = FileStatsHandler()
    seqid = repserv.apply_diffs(h, seqid, maxkb)
    print("Final sequence id:", seqid)

    h.nodes.outstats("Nodes")
    h.ways.outstats("Ways")
    h.rels.outstats("Relations")



"""
Extract objects with an amenity tag from an osm file and list them
with their name and position.

Points and areas will be processed, unclosed ways will be skipped.

This example shows how geometries from osmium objects can be imported
into shapely using the WKBFactory.
"""
import osmium
import sys
import shapely.wkb as wkblib

wkbfab = osmium.geom.WKBFactory()


class AmenityListHandler(osmium.SimpleHandler):

    def print_amenity(self, tags, lon, lat):
        name = tags.get('name', '')
        print("%f %f %-15s %s" % (lon, lat, tags['amenity'], name))

    def node(self, n):
        self.print_amenity(n.tags, n.location.lon, n.location.lat)

    def area(self, a):
        wkb = wkbfab.create_multipolygon(a)
        poly = wkblib.loads(wkb, hex=True)
        centroid = poly.representative_point()
        self.print_amenity(a.tags, centroid.x, centroid.y)


def main(osmfile):

    handler = AmenityListHandler()

    handler.apply_file(osmfile, filters=[osmium.filter.KeyFilter('amenity')])

    return 0


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: python %s <osmfile>" % sys.argv[0])
        sys.exit(-1)

    exit(main(sys.argv[1]))



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Tuple, Iterable

from ._osmium import BaseFilter
from .osm import osm_entity_bits

class EmptyTagFilter(BaseFilter):
    """ Filter class which only lets pass objects which have at least one tag.
    """
    def __init__(self) -> None:
        """ Create a new filter object.
        """


class KeyFilter(BaseFilter):
    """ Filter class which lets objects pass which have tags with at
        least one of the listed keys.

        This filter functions like an OR filter. To create an AND filter
        (a filter that lets object pass that have tags with all the listed
        keys) you need to chain multiple KeyFilter objects.
    """
    def __init__(self, *keys: str) -> None:
        """ Create a new filter object. The parameters list the keys
            by which the filter should choose the objects. At least one
            key is required.
        """


class TagFilter(BaseFilter):
    """ Filter class which lets objects pass which have tags with at
        least one of the listed key-value pair.

        This filter functions like an OR filter. To create an AND filter
        (a filter that lets object pass that have tags with all the listed
        key-value pairs) you need to chain multiple TagFilter objects.
    """
    def __init__(self, *tags: Tuple[str, str]) -> None:
        """ Create a new filter object. The parameters list the key-value
            pairs by which the filter should choose objects. Each pair must
            be a tuple with two strings and at least one pair is required.
        """


class EntityFilter(BaseFilter):
    """ Filter class which lets pass objects according to their type.
    """
    def __init__(self, entities: osm_entity_bits) -> None:
        """ Crate a new filter object. Only objects whose type is listed
            in _entities_ can pass the filter.
        """


class IdFilter(BaseFilter):
    """ Filter class which only lets pass objects with given IDs.

        This filter usually only makes sense when used together with
        a type restriction, set using `enable_for()`.
    """
    def __init__(self, ids: Iterable[int]) -> None:
        """ Create a new filter object. _ids_ contains the IDs the filter
            should let pass. It can be any iterable over ints.
        """


class GeoInterfaceFilter(BaseFilter):
    """ Filter class, which adds a [__geo_interface__](https://gist.github.com/sgillies/2217756) attribute to object which have geometry information.

        The filter can process node, way and area types. All other types
        will be dropped. To create geometries for ways, the location cache needs
        to be enabled. Relations and closed ways can only be transformed
        to polygons when the area handler is enabled.
    """
    def __init__(self, drop_invalid_geometries: bool= ..., tags: Iterable[str] = ...) -> None:
        """ Create a new filter object. The filter will usually drop all
            objects that do not have a geometry. Set _drop_invalid_geometries_
            to `False` to just let them pass.

            The filter will normally add all tags it finds as properties to
            the GeoInterface output. To filter the tags to relevant ones, set
            _tags_ to the desired list.
        """



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest
import uuid

import osmium
from helpers import IDCollector


@pytest.mark.parametrize('init', [None, 1])
def test_file_processor_bad_init(init):
    with pytest.raises(TypeError):
        for obj in osmium.FileProcessor(init):
            pass


def test_simple_generator(opl_buffer):
    count = 0
    for obj in osmium.FileProcessor(opl_buffer('n1 x5 y5')):
        assert obj.type_str() == 'n'
        assert obj.id == 1
        count += 1

    assert count == 1


def test_generator_with_location(opl_buffer):
    data = opl_buffer("""\
              n1 x10 y20
              n2 x11 y21
              w45 Nn1,n2
              """)

    count = 0
    for obj in osmium.FileProcessor(data).with_locations():
        count += 1
        if obj.type_str() == 'w':
            assert len(obj.nodes) == 2
            assert [n.ref for n in obj.nodes] == [1, 2]
            assert [n.location.lon for n in obj.nodes] == [10, 11]
            assert [n.location.lat for n in obj.nodes] == [20, 21]

    assert count == 3


def test_generator_with_areas(opl_buffer):
    data = opl_buffer("""\
            n10 x3 y3
            n11 x3 y3.01
            n12 x3.01 y3.01
            n13 x3.01 y3
            w12 Nn10,n11,n12,n13,n10 Tbuilding=yes
            """)

    count = 0
    for obj in osmium.FileProcessor(data).with_areas():
        if obj.type_str() == 'a':
            count += 1
            assert obj.from_way()
            assert obj.orig_id() == 12

    assert count == 1


def test_generator_with_areas_with_filter(opl_buffer):
    data = opl_buffer("""\
            n10 x3 y3
            n11 x3 y3.01
            n12 x3.01 y3.01
            n13 x3.01 y3
            w11 Nn10,n11,n12,n13
            w12 Nn13,n10
            r1 Tbuilding=yes,type=multipolygon Mw11@,w12@
            """)

    count = 0
    for obj in osmium.FileProcessor(data)\
                     .with_areas()\
                     .with_filter(osmium.filter.EntityFilter(osmium.osm.AREA)):
        assert obj.is_area()
        count += 1

    assert count == 1


def test_generator_with_areas_with_area_filter(opl_buffer):
    data = opl_buffer("""\
            n10 x3 y3
            n11 x3 y3.01
            n12 x3.01 y3.01
            n13 x3.01 y3
            w11 Nn10,n11,n12,n13
            w12 Nn13,n10
            r1 Tbuilding=yes,type=multipolygon Mw11@,w12@
            r2 Tlanduse=grass,type=multipolygon Mw11@,w12@
            """)

    count = 0
    for obj in osmium.FileProcessor(data)\
                     .with_areas(osmium.filter.KeyFilter('building'))\
                     .with_filter(osmium.filter.EntityFilter(osmium.osm.AREA)):
        assert obj.is_area()
        assert obj.id == 3
        count += 1

    assert count == 1


def test_generator_with_filter(opl_buffer):
    data = opl_buffer("""\
            n10 x3 y3
            n11 x3 y3.01 Tfoo=bar
            """)

    count = 0
    for obj in osmium.FileProcessor(data).with_filter(osmium.filter.EmptyTagFilter()):
        count += 1
        assert obj.type_str() == 'n'
        assert obj.id == 11

    assert count == 1


def test_file_processor_header(tmp_path):
    fn = tmp_path / f"{uuid.uuid4()}.xml"
    fn.write_text("""<?xml version='1.0' encoding='UTF-8'?>
    <osm version="0.6" generator="test-pyosmium" timestamp="2014-08-26T20:22:02Z">
         <bounds minlat="-90" minlon="-180" maxlat="90" maxlon="180"/>
    </osm>
    """)

    h = osmium.FileProcessor(fn).header

    assert not h.has_multiple_object_versions
    assert h.box().valid()
    assert h.box().size() == 64800.0


def test_file_processor_access_nodestore(opl_buffer):
    fp = osmium.FileProcessor(opl_buffer('n56 x3 y-3'))\
          .with_locations(osmium.index.create_map('sparse_mem_map'))

    for _ in fp:
        pass

    assert fp.node_location_storage.get(56).lat == -3
    assert fp.node_location_storage.get(56).lon == 3


def test_file_processor_bad_location_type(opl_buffer):
    with pytest.raises(TypeError, match='LocationTable'):
        osmium.FileProcessor(opl_buffer('n56 x3 y-3')).with_locations(67)


def test_propagate_data_from_filters(opl_buffer):
    class MyFilter:
        def node(self, n):
            n.saved = 'test'
            return False

    fp = osmium.FileProcessor(opl_buffer('n56 x3 y-3')).with_filter(MyFilter())

    for obj in fp:
        assert obj.saved == 'test'


def test_simple_zip(opl_buffer):
    fp1 = osmium.FileProcessor(opl_buffer("""\
            n1
            n3
            n5
            w10 Nn1,n2
            r1 Mw1@
          """))

    fp2 = osmium.FileProcessor(opl_buffer("""\
            n2
            n3
            n456
            w12 Nn45,n85
            r1 Mw2@
          """))

    results = []
    for o1, o2 in osmium.zip_processors(fp1, fp2):
        results.append(((None if o1 is None else o1.type_str() + str(o1.id)),
                       (None if o2 is None else o2.type_str() + str(o2.id))))

    assert results == [('n1', None),
                       (None, 'n2'),
                       ('n3', 'n3'),
                       ('n5', None),
                       (None, 'n456'),
                       ('w10', None),
                       (None, 'w12'),
                       ('r1', 'r1')]


def test_filtered_handler_python(opl_buffer):
    data = opl_buffer("""\
            n1 Tamenity=foo
            n3
            w1 Thighway=residential
            w2
            r4
            """)

    ids = IDCollector()

    processed = []

    fp = osmium.FileProcessor(data)\
               .handler_for_filtered(ids)\
               .with_filter(osmium.filter.EmptyTagFilter())

    for obj in fp:
        processed.append(f"{obj.type_str()}{obj.id}")

    assert processed == ['n1', 'w1']
    assert ids.nodes == [3]
    assert ids.ways == [2]
    assert ids.relations == [4]


def test_filtered_handler_basehandler(opl_buffer, tmp_path):
    data = opl_buffer("""\
            n1 Tamenity=foo
            n3
            w1 Thighway=residential
            w2
            r4
            """)

    testf = tmp_path / f"{uuid.uuid4()}.opl"

    with osmium.SimpleWriter(str(testf)) as writer:
        fp = osmium.FileProcessor(data)\
                .handler_for_filtered(writer)\
                .with_filter(osmium.filter.EmptyTagFilter())

        processed = []
        for obj in fp:
            processed.append(f"{obj.type_str()}{obj.id}")

    assert processed == ['n1', 'w1']

    ids = IDCollector()

    osmium.apply(testf, ids)

    assert ids.nodes == [3]
    assert ids.ways == [2]
    assert ids.relations == [4]



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from pathlib import Path

import pytest
import osmium

pytestmark = pytest.mark.iterations(1)

TEST_DIR = (Path(__file__) / '..').resolve()


class DanglingReferenceBase:
    """ Base class for tests that try to keep a reference to the object
        that was handed into the callback. We expect that the handler
        bails out with a runtime error in such a case.
    """
    node = None
    way = None
    relation = None
    area = None
    refkeeper = []

    def keep(self, obj, func):
        self.refkeeper.append((obj, func))

    def test_keep_reference(self):
        h = osmium.make_simple_handler(node=self.node, way=self.way,
                                       relation=self.relation, area=self.area)
        h.apply_file(TEST_DIR / 'example-test.osc')
        assert self.refkeeper

        for obj, func in self.refkeeper:
            with pytest.raises(RuntimeError, match="removed OSM object"):
                func(obj)
            # str() and repr() must not throw errors
            str(obj)
            repr(obj)

    def test_keep_reference_generator(self):
        for obj in osmium.FileProcessor(TEST_DIR / 'example-test.osc').with_areas():
            if obj.type_str() == 'n' and self.node is not None:
                self.node(obj)
            elif obj.type_str() == 'w' and self.way is not None:
                self.way(obj)
            elif obj.type_str() == 'r' and self.relation is not None:
                self.relation(obj)
            elif obj.type_str() == 'a' and self.area is not None:
                self.area(obj)

        assert self.refkeeper

        for obj, func in self.refkeeper:
            with pytest.raises(RuntimeError, match="removed OSM object"):
                func(obj)
            # str() and repr() must not throw errors
            str(obj)
            repr(obj)


class TestKeepNodeRef(DanglingReferenceBase):

    def node(self, n):
        self.keep(n, lambda n: n.id)


class TestKeepWayRef(DanglingReferenceBase):

    def way(self, w):
        self.keep(w, lambda n: n.id)


class TestKeepRelationRef(DanglingReferenceBase):

    def relation(self, r):
        self.keep(r, lambda n: n.id)


class TestKeepAreaRef(DanglingReferenceBase):

    def area(self, a):
        self.keep(a, lambda n: n.id)


class TestKeepNodeTagsRef(DanglingReferenceBase):

    def node(self, n):
        self.keep(n.tags, lambda t: 'foo' in t)


class TestKeepWayTagsRef(DanglingReferenceBase):

    def way(self, w):
        self.keep(w.tags, lambda t: 'foo' in t)


class TestKeepRelationTagsRef(DanglingReferenceBase):

    def relation(self, r):
        self.keep(r.tags, lambda t: 'foo' in t)


class TestKeepAreaTagsRef(DanglingReferenceBase):

    def area(self, a):
        self.keep(a.tags, lambda t: 'foo' in t)


class TestKeepTagListIterator(DanglingReferenceBase):

    def node(self, n):
        self.keep(n.tags.__iter__(), lambda t: next(t))


class TestKeepOuterRingIterator(DanglingReferenceBase):

    def area(self, r):
        self.keep(r.outer_rings(), lambda t: next(t))


class TestKeepOuterRing(DanglingReferenceBase):

    def area(self, r):
        for ring in r.outer_rings():
            self.keep(ring, lambda t: len(t))


class TestKeepInnerRingIterator(DanglingReferenceBase):

    def area(self, r):
        for ring in r.outer_rings():
            self.keep(r.inner_rings(ring), lambda t: next(t))


class TestKeepInnerRing(DanglingReferenceBase):

    def area(self, r):
        for outer in r.outer_rings():
            for inner in r.inner_rings(outer):
                self.keep(inner, lambda t: len(t))


class TestKeepRelationMemberIterator(DanglingReferenceBase):

    def relation(self, r):
        self.keep(r.members, lambda t: next(t))


class NotADanglingReferenceBase:
    """ Base class for tests that ensure that the callback does not
        bail out because of dangling references when POD types are
        kept.
    """

    node = None
    way = None
    relation = None
    area = None
    refkeeper = []

    def keep(self, obj, func):
        self.refkeeper.append((obj, func))

    def test_keep_reference(self):
        h = osmium.make_simple_handler(node=self.node, way=self.way,
                                       relation=self.relation, area=self.area)
        h.apply_file(TEST_DIR / 'example-test.pbf')
        assert self.refkeeper

        for obj, func in self.refkeeper:
            func(obj)

    def test_keep_reference_generator(self):
        for obj in osmium.FileProcessor(TEST_DIR / 'example-test.pbf').with_areas():
            if obj.is_node() and self.node is not None:
                self.node(obj)
            elif obj.is_way() and self.way is not None:
                self.way(obj)
            elif obj.is_relation() and self.relation is not None:
                self.relation(obj)
            elif obj.is_area() and self.area is not None:
                self.area(obj)

        assert self.refkeeper

        for obj, func in self.refkeeper:
            func(obj)


class TestKeepLocation(NotADanglingReferenceBase):

    def node(self, n):
        self.keep(n.location, lambda loc: loc.x)


class TestKeepNode(NotADanglingReferenceBase):

    def node(self, n):
        for t in n.tags:
            self.keep(t, lambda t: t.k)


class TestKeepMember(NotADanglingReferenceBase):

    def relation(self, r):
        for m in r.members:
            self.keep(m, lambda m: m.ref)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import json

import pytest
import osmium.geom


@pytest.fixture
def node_geom(test_data):

    def _run(factory, data='n1 x-23.3 y28.0'):
        geoms = []

        def _mk_point(node):
            geoms.append(factory.create_point(node))
        handler = osmium.make_simple_handler(node=_mk_point)
        handler.apply_file(test_data(data))

        assert len(geoms) == 1
        return geoms[0]

    return _run


def test_wkb_create_node(node_geom):
    wkb = node_geom(osmium.geom.WKBFactory())
    if wkb.startswith('01'):
        assert wkb.startswith('0101000000')
    else:
        assert wkb.startswith('00')


def test_wkt_create_node(node_geom):
    wkt = node_geom(osmium.geom.WKTFactory())
    assert wkt.startswith('POINT(')


def test_geojson_create_node(node_geom):
    geom = node_geom(osmium.geom.GeoJSONFactory())
    geom = json.loads(geom)
    assert geom['type'], 'Point'


@pytest.fixture
def way_geom(test_data):

    def _run(factory):
        opl = test_data(['n1 x0 y0', 'n2 x1 y0', 'n3 x0 y1', 'w1 Nn1,n2,n3'])
        geoms = []

        def _mk_way(w):
            geoms.append(factory.create_linestring(w))
            geoms.append(factory.create_linestring(
                w, direction=osmium.geom.direction.BACKWARD))
            geoms.append(factory.create_linestring(
                w, use_nodes=osmium.geom.use_nodes.ALL))

        handler = osmium.make_simple_handler(way=_mk_way)
        handler.apply_file(opl, locations=True)

        assert len(geoms) == 3
        return geoms

    return _run


def test_wkb_create_way(way_geom):
    wkbs = way_geom(osmium.geom.WKBFactory())

    for wkb in wkbs:
        if wkb.startswith('01'):
            assert wkb.startswith('0102000000030'), "wkb: " + wkb
        else:
            assert wkb.startswith('00')


def test_wkt_create_way(way_geom):
    wkts = way_geom(osmium.geom.WKTFactory())

    assert all(wkt.startswith('LINESTRING(') for wkt in wkts)


def test_geojson_create_way(way_geom):
    geoms = way_geom(osmium.geom.GeoJSONFactory())

    assert all(json.loads(geom)['type'] == 'LineString' for geom in geoms)


@pytest.fixture
def area_geom(test_data):

    def _run(factory):
        opl = test_data(['n1 x0 y0', 'n2 x1 y0', 'n3 x0 y1', 'w23 Nn1,n2,n3,n1 Tarea=yes'])
        geoms = []

        def _mk_area(a):
            geoms.append(factory.create_multipolygon(a))

        handler = osmium.make_simple_handler(area=_mk_area)
        handler.apply_file(opl, locations=True)

        assert len(geoms) == 1
        return geoms[0]

    return _run


def test_wkb_create_poly(area_geom):
    wkb = area_geom(osmium.geom.WKBFactory())
    if wkb.startswith('01'):
        assert wkb.startswith('010600000001'), "wkb: " + wkb
    else:
        assert wkb.startswith('00')


def test_wkt_create_poly(area_geom):
    wkt = area_geom(osmium.geom.WKTFactory())
    assert wkt.startswith('MULTIPOLYGON(')


def test_geojson_create_poly(area_geom):
    geom = area_geom(osmium.geom.GeoJSONFactory())
    geom = json.loads(geom)
    assert geom['type'] == 'MultiPolygon'


def test_lonlat_to_mercator():
    c = osmium.geom.lonlat_to_mercator(osmium.geom.Coordinates(3.4, -7.3))
    assert c.x == pytest.approx(378486.2686971)
    assert c.y == pytest.approx(-814839.8325696)


def test_mercator_lonlat():
    c = osmium.geom.mercator_to_lonlat(osmium.geom.Coordinates(0.03, 10.2))
    assert c.x == pytest.approx(0.00000026, rel=1e-1)
    assert c.y == pytest.approx(0.00009162, rel=1e-1)


def test_coordinate_from_location():
    c = osmium.geom.Coordinates(osmium.osm.Location(10.0, -3.0))
    assert c.x == pytest.approx(10.0)
    assert c.y == pytest.approx(-3.0)


def test_haversine():
    data = ['n1 x0 y0', 'n2 x1 y0', 'n3 x0 y1', 'w1 Nn1,n2,n3']

    results = []

    def call_haversine(w):
        results.append(osmium.geom.haversine_distance(w.nodes))

    handler = osmium.make_simple_handler(way=call_haversine)
    handler.apply_buffer('\n'.join(data).encode('utf-8'), 'opl', locations=True)

    assert 1 == len(results)
    assert 268520 == pytest.approx(results[0])


def test_haversine_invalid_object():
    data = ['n1 x0 y0', 'n2 x1 y0', 'n3 x0 y1', 'w1 Nn1,n2,n3']

    results = []

    def call_haversine(w):
        results.append(w.nodes)
    handler = osmium.make_simple_handler(way=call_haversine)
    handler.apply_buffer('\n'.join(data).encode('utf-8'), 'opl', locations=True)

    assert results

    with pytest.raises(RuntimeError, match="removed OSM object"):
        osmium.geom.haversine_distance(results[0])


def test_haversine_coordinates():
    dist = osmium.geom.haversine_distance(osmium.geom.Coordinates(0, 0),
                                          osmium.geom.Coordinates(1, 1))
    assert dist == pytest.approx(157293.74877)


def test_haversine_location():
    dist = osmium.geom.haversine_distance(osmium.osm.Location(0, 0),
                                          osmium.osm.Location(1, 1))
    assert dist == pytest.approx(157293.74877)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from pathlib import Path

import pytest
import osmium

from helpers import CountingHandler


TEST_FILE = str((Path(__file__) / '..' / 'example-test.pbf').resolve())


@pytest.mark.thread_unsafe
def test_area_handler():
    area = osmium.area.AreaManager()

    osmium.apply(osmium.io.Reader(TEST_FILE), area.first_pass_handler())

    ch_area = CountingHandler()
    ch_others = CountingHandler()

    lh = osmium.NodeLocationsForWays(osmium.index.create_map("flex_mem"))
    lh.ignore_errors()

    osmium.apply(osmium.io.Reader(TEST_FILE), lh,
                 ch_others, area.second_pass_handler(ch_area))

    assert ch_area.counts == [0, 0, 0, 5239]
    assert ch_others.counts == [211100, 10315, 244, 0]


def test_area_buffer_handler():
    area = osmium.area.AreaManager()

    osmium.apply(osmium.io.Reader(TEST_FILE), area.first_pass_handler())

    lh = osmium.NodeLocationsForWays(osmium.index.create_map("flex_mem"))
    lh.ignore_errors()

    buf = osmium.BufferIterator()

    osmium.apply(osmium.io.Reader(TEST_FILE), lh, area.second_pass_to_buffer(buf))

    counts = 0
    for obj in buf:
        assert obj.is_area()
        counts += 1

    assert counts == 5239
